/* gifout.c by Goran */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "print.h"

#ifdef __MINGW32__
#include <windows.h>
#endif

static FILE *fout=NULL;
static char *fname=NULL;
static int printModeType;

static int heightInPixels;
static int Done;		/* IO flag */

#define JFWRITE(file,buf,sizeofbuf)  \
  ((size_t) fwrite((void *) (buf), (size_t) 1, (size_t) (sizeofbuf), (file)))
#define	MAX_LZW_BITS	12	/* maximum LZW code size (4096 symbols) */
#define LZW_TABLE_SIZE	((code_int) 1 << MAX_LZW_BITS)
#define HSIZE		5003	/* hash table size for 80% occupancy */
#define MAXCODE(n_bits)	(((code_int) 1 << (n_bits)) - 1)

typedef short code_int;		/* must hold -1 .. 2**MAX_LZW_BITS */
typedef int hash_int;		/* must hold -2*HSIZE..2*HSIZE */

static int n_bits;		/* current number of bits/code */
static code_int maxcode;	/* maximum code, given n_bits */
static int init_bits;		/* initial n_bits ... restored after clear */
static code_int ClearCode;	/* clear code (doesn't change) */
static code_int EOFCode;	/* EOF code (ditto) */
static code_int free_code;	/* first not-yet-used symbol code */
static code_int *hash_code=NULL;/* => hash table of symbol codes */
static code_int *hash_prefix=NULL;/* => hash table of prefix symbols */
static char *hash_suffix=NULL;	/* => hash table of suffix bytes */
static int bytesinpkt;		/* # of bytes in current packet */
static char packetbuf[256];	/* workspace for accumulating packet */
static int cur_accum;		/* holds bits not yet output */
static int cur_bits;		/* # of bits in cur_accum */
static code_int waiting_code;	/* symbol not yet output; may be extendable */
static int first_byte;		/* if TRUE, waiting_code is not valid */

static void flush_packet(void);
static void char_out(char c);
static void output(code_int code);
static void clear_hash(void);
static void clear_block(void);
static void compress_init(int i_bits);
static void compress_byte(int c);
static void compress_term(void);
static void put_word(unsigned short w);

/*
 * The LZW hash table consists of three parallel arrays:
 *   hash_code[i]	code of symbol in slot i, or 0 if empty slot
 *   hash_prefix[i]	symbol's prefix code; undefined if empty slot
 *   hash_suffix[i]	symbol's suffix character; undefined if empty slot
 * where slot values (i) range from 0 to HSIZE-1.
 *
 * Algorithm:  use open addressing double hashing (no chaining) on the
 * prefix code / suffix character combination.  We do a variant of Knuth's
 * algorithm D (vol. 3, sec. 6.4) along with G. Knott's relatively-prime
 * secondary probe. */

/* Routines to package compressed data bytes into GIF data blocks.
 * A data block consists of a count byte (1..255) and that many data bytes.*/
static void flush_packet(void)
/* flush any accumulated data */
{
  if (bytesinpkt > 0) 
  {
    /* never write zero-length packet */
    packetbuf[0] = (char) bytesinpkt++;
    Done = (JFWRITE(fout, packetbuf, bytesinpkt) == (size_t) bytesinpkt);
    bytesinpkt = 0;
  }
 fflush(fout);
}

static void char_out(char c)
/* Add a character to current packet; flush to disk if necessary */
{
  packetbuf[++bytesinpkt] = c;
  if (bytesinpkt >= 255)
    flush_packet();
}

/* Routine to convert variable-width codes into a byte stream */
static void output(code_int code)
/* Emit a code of n_bits bits
 * Uses cur_accum and cur_bits to reblock into 8-bit bytes */
{
  if (cur_bits > 0)
    cur_accum |= ((int) code << cur_bits);
  else
    cur_accum = code;
  cur_bits += n_bits;
  while (cur_bits >= 8) 
  {
    char_out((char) (cur_accum & 0xFF));
    cur_accum >>= 8;
    cur_bits -= 8;
  }
  /* If the next entry is going to be too big for the code size,
   * then increase it, if possible.  We do this here to ensure
   * that it's done in sync with the decoder's codesize increases.*/
  if (free_code > maxcode) 
  {
    n_bits++;
    if (n_bits == MAX_LZW_BITS)
      maxcode = LZW_TABLE_SIZE;	  /* free_code will never exceed this */
    else
      maxcode = MAXCODE(n_bits);
  }
}

/* The LZW algorithm proper */
static void clear_hash(void)
/* Fill the hash table with empty entries */
{
  int i;	
  
  /* It's sufficient to zero hash_code[] */
  for( i=0; i<HSIZE; i++ )
    hash_code[i] = (code_int) 0;
}

static void clear_block(void)
/* Reset compressor and issue a Clear code */
{
  clear_hash();			/* delete all the symbols */
  free_code = ClearCode + 2;
  output(ClearCode);		/* inform decoder */
  n_bits = init_bits;		/* reset code size */
  maxcode = MAXCODE(n_bits);
}

static void compress_init(int i_bits)
/* Initialize LZW compressor */
{
  /* init all the static variables */
  n_bits = init_bits = i_bits;
  maxcode = MAXCODE(n_bits);
  ClearCode = ((code_int) 1 << (init_bits - 1));
  EOFCode = ClearCode + 1;
  free_code = ClearCode + 2;
  first_byte = TRUE;	/* no waiting symbol yet */
  /* init output buffering vars */
  bytesinpkt = 0;
  cur_accum = 0;
  cur_bits = 0;
  /* clear hash table */
  clear_hash();
  /* GIF specifies an initial Clear code */
  output(ClearCode);
}

static void compress_byte(int c)
/* Accept and compress one 8-bit byte */
{
  hash_int i;
  hash_int disp;

  if (first_byte) 
  {		
    /* need to initialize waiting_code */
    waiting_code = (code_int) c;
    first_byte = FALSE;
    return;
  }
  /* Probe hash table to see if a symbol exists for
   * waiting_code followed by c.
   * If so, replace waiting_code by that symbol and return.*/
  i = ((hash_int) c << (MAX_LZW_BITS-8)) + waiting_code;
  /* i is less than twice 2**MAX_LZW_BITS, therefore less than twice HSIZE */
  if (i >= HSIZE) i -= HSIZE;
  if (hash_code[i] != 0) 
  {	
    /* is first probed slot empty? */
    if (hash_prefix[i] == waiting_code && hash_suffix[i] == (char) c) 
    {
      waiting_code = hash_code[i];
      return;
    }
    if (i == 0)	   /* secondary hash (after G. Knott) */
      disp = 1;
    else
      disp = HSIZE - i;
    while (1) 
    {
      i -= disp;
      if (i < 0)
	i += HSIZE;
      if (hash_code[i] == 0)
	break;			/* hit empty slot */
      if (hash_prefix[i] == waiting_code && hash_suffix[i] == (char) c) 
      {
	waiting_code = hash_code[i];
	return;
      }
    }
  }
  /* here when hashtable[i] is an empty slot; desired symbol not in table */
  output(waiting_code);

  if (free_code < LZW_TABLE_SIZE) 
   {
    hash_code[i] = free_code++;	  /* add symbol to hashtable */
    hash_prefix[i] = waiting_code;
    hash_suffix[i] = (char) c;
   }
  else
   clear_block();
  waiting_code = (code_int) c;
}

/* Clean up at end */
static void compress_term()
{
 /* Flush out the buffered code */
 if (! first_byte)
   output(waiting_code);
 /* Send an EOF code */
 output(EOFCode);
 /* Flush the bit-packing buffer */
 if (cur_bits > 0) 
 {
   char_out((char) (cur_accum & 0xFF));
 }
 /* Flush the packet buffer */
 flush_packet();
}

/* GIF header construction */

/* Emit a 16-bit word, LSB first */
static void put_word(unsigned short w)
{
 putc(w & 0xFF, fout);
 putc((w >> 8) & 0xFF, fout);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
 int lines, char bandType, unsigned char *cmap)
{
 static int blackx, whitex;
 int x, y;
 static int InitCodeSize;

 if (data == NULL)
  return(PRT_INTERNAL_ERROR);

 if (bandType > printModeType)
  return(PRT_WRONG_IMG);

 if (ypos == 0)
  {
   if (bandType == IMG_BW)
    {
     int i, m, blackM, whiteM;

     blackx = whitex = 0;
     blackM = whiteM = cmap[0] + cmap[1] + cmap[2];
     for (i=1; i < 256; i++)
     {
      m = cmap[i*3] + cmap[i*3 + 1] + cmap[i*3 + 2];
      if (blackM > m)
       { blackM = m; blackx = i; }
      else if (whiteM < m)
       { whiteM = m; whitex = i; }
     }
     if (blackx == whitex)
      {
       blackx = 0; cmap[0] = cmap[1] = cmap[2] = (unsigned char) 0;
       whitex = 1; cmap[3] = cmap[4] = cmap[5] = (unsigned char) 255;
      }
    }
   InitCodeSize=(printModeType == IMG_BW) ? 2 : 8; /* 8-bpp cmap image */

   /* Write the Global Color Map */
   if (printModeType == IMG_BW)
    Done = (JFWRITE(fout, cmap, 2*3) == 2*3);
   else
    Done = (JFWRITE(fout, cmap, 256*3) == 256*3);
   if (!Done)
    return(PRT_IO_ERROR);

   /* Write image separator and Image Descriptor */
   putc(',', fout); /* separator */
   put_word((unsigned short) 0); /* left/top offset */
   put_word((unsigned short) 0);
   put_word((unsigned short) xpixels);   /* image size */
   put_word((unsigned short) heightInPixels);
   /* flag byte: not interlaced, no local color map */
   putc(0, fout);

   /* Write Initial Code Size byte */
   putc(InitCodeSize, fout);

   /* Initialize for LZW compression of image data */
   compress_init(InitCodeSize+1);
  }

 if (bandType == IMG_BW)
  {
   for (y = 0; y < lines; y++)
    {
     int i, xbytes, remain;
     unsigned char databyte;

     xbytes = xpixels / 8;
     remain = xpixels % 8;
     for (x = 0; x < xbytes; x++)
      {
       databyte = *data++;
       for (i = 7; i >= 0 ; i--)
	compress_byte((databyte & (1<<i)) ? blackx : whitex);
      }
     if (remain > 0)
      {
       databyte = *data;
       for (i = 7; i >= 8 - remain; i--)
	compress_byte((databyte & (1<<i)) ? blackx : whitex);
      }
     data += bpl - xbytes;
    }
  }
 else
  for (y = 0; y < lines; y++)
  {
   for (x = 0; x < xpixels; x++)
     compress_byte((int) ((unsigned char) *data++));
   data += bpl - xpixels;
  }
  
 if (Done)
  return(PRT_OK);
 else
  return(PRT_IO_ERROR);
}

static int initPrint(int printType, int complexImage, int xpixels,
	int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 heightInPixels = ypixels;
 int filedes = -1;

 printModeType = printType;
 if (printModeType > IMG_TRUE_COL_DITH)
  return(PRT_WRONG_IMG);

 if (fname)
  free(fname);
 if (fout)
  fclose(fout);
#ifdef __MINGW32__
 char tempPath[MAX_PATH + 1];

 if (GetTempPathA(MAX_PATH + 1, tempPath) == 0)
  {
   // Failed to get temp path.
   perror("plib: failed to get temp path");
   return (PRT_IO_ERROR);
  }

 // This is allocated in advance because of the code below.
 char tempFilename[MAX_PATH + 1];
 fname = tempFilename;
 if (GetTempFileNameA(tempPath, "prt", 1, fname) == 0)
  {
   perror("plib: failed to create tmpfile");
   return(PRT_IO_ERROR);
  }

 fout = fopen(fname, "wb");
 fname = strdup(fname);
#else
 fname = strdup("/tmp/prtXXXXXX");
 if ((filedes=mkstemp(fname)) < 0 )
  return(PRT_IO_ERROR);
 else
  {
   /* instead of fout = fdopen(filedes, "wb"); */
   /* close and reopen the file as fclose does not close files created by mkstemp */
   close(filedes);
   fout = fopen(fname, "wb");
  }
#endif

 if (hash_code)
  free(hash_code);
 hash_code = (code_int *) malloc(HSIZE * sizeof(code_int));
 if (hash_code == NULL)
  return(PRT_NO_MEMORY);
 if (hash_prefix)
  free(hash_prefix);
 hash_prefix = (code_int *) malloc(HSIZE * sizeof(code_int));
 if (hash_prefix == NULL)
  return(PRT_NO_MEMORY);
 if (hash_suffix)
  free(hash_suffix);
 hash_suffix = (char *) malloc(HSIZE * sizeof(char));
 if (hash_suffix == NULL)
  return(PRT_NO_MEMORY);

 Done = (JFWRITE(fout, "GIF87a", 6) == 6);
 if (!Done)
  return(PRT_IO_ERROR);

 /* Write the Logical Screen Descriptor */
 put_word((unsigned short) xpixels);
 put_word((unsigned short) ypixels);
 if (printModeType==IMG_BW)
  putc(0x80/*0xF7*/, fout);
 else
  putc(0xF7, fout); /* 8 bits per pixel */
 putc(0, fout);	/* Background color index */
 putc(0, fout);	/* Reserved in GIF87 (aspect ratio in GIF89) */

 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 if (printStatus == PRT_OK)
  if (printModeType <= IMG_TRUE_COL_DITH)
   {
    compress_term();
    if (!Done)
     printStatus=PRT_IO_ERROR;
    if (fout)
     {
      putc(0, fout);
      putc(';', fout);
      fflush(fout);
      fclose(fout);
      fout=NULL;
     }
    if (fname)
     printStatus = enqueuePrintFile(fname);
   }
  else 
   printStatus = PRT_WRONG_IMG;
 else
  if (printModeType <= IMG_TRUE_COL_DITH)
   {
    if (fout)
     {
      fclose(fout);
      fout=NULL;
     }
   }
  else
   printStatus = PRT_WRONG_IMG;

 if (hash_code)
  { free(hash_code); hash_code=NULL; }
 if (hash_prefix)
  { free(hash_prefix); hash_prefix=NULL; }
 if (hash_suffix)
  { free(hash_suffix); hash_suffix=NULL; }

 if (fname)
  {
   removeFileFreeFilename(fname, printStatus);
   fname=NULL;
  }
 return(printStatus);
}

/* fill in the preallocated printer modes for GIF */
static void getPrtInfo(printerMode *printerMod, int *numModes,
	printerDensity *prtDensityX, int *numDensitX,
	printerDensity *prtDensityY, int *numDensitY,
	int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
	int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_CMAP;
 printerMod[1].printType = IMG_BW;

 *numModes=2;
 *numDensitX = *numDensitY = 7;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = MAX_PRINT;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }
 prtDensityX[6].pixels = prtDensityY[6].pixels = MAX_PRINT;
 prtDensityX[6].dpi = prtDensityY[6].dpi = 720000;

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=TRUE;
 *prtName="GIF";
 *profileName=NULL;
}

/* register printing to GIF file */
void registerGIF()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}
