/* ppng.c */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <png.h>
#include "print.h"

#ifdef __MINGW32__
#include <windows.h>
#endif

static FILE *fp=NULL;
static char *fname=NULL;
static int printModeType;
static png_structp png_ptr;
static png_infop info_ptr;
static int xpix, ypix;
static png_colorp palette=NULL;

static int initPrint(int printType, int complexImage, int xpixels,
 int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 png_uint_32 res_x, res_y;
 int filedes = -1;

 printModeType = printType;
 if (fname)
  free(fname);
#ifdef __MINGW32__
 char tempPath[MAX_PATH + 1];

 if (GetTempPathA(MAX_PATH + 1, tempPath) == 0)
  {
   // Failed to get temp path.
   perror("plib: failed to get temp path");
   return (PRT_IO_ERROR);
  }

 // This is allocated in advance because of the code below.
 char tempFilename[MAX_PATH + 1];
 fname = tempFilename;
 if (GetTempFileNameA(tempPath, "prt", 1, fname) == 0)
  {
   perror("plib: failed to create tmpfile");
   return(PRT_IO_ERROR);
  }

 fp = fopen(fname, "wb");
 fname = strdup(fname);
#else
 fname = strdup("/tmp/prtXXXXXX");
 if ((filedes=mkstemp(fname)) < 0 )
  return(PRT_IO_ERROR);
 else
  {
   /* instead of fp = fdopen(filedes, "wb"); */
   /* close and reopen the file as fclose does not close files created by mkstemp */
   close(filedes);
   fp = fopen(fname, "wb");
  }
#endif

 png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
    (void *) NULL, NULL, NULL);
 if (!png_ptr)
  { fclose(fp); return(PRT_NO_MEMORY); }

 info_ptr = png_create_info_struct(png_ptr);
 if (!info_ptr)
  {
   fclose(fp);
   png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
   return(PRT_NO_MEMORY);
  }

// if (setjmp(png_ptr->jmpbuf))
 if (setjmp(png_jmpbuf(png_ptr)))
  {
   fclose(fp);
   png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
   return(PRT_IO_ERROR);
  }

 png_init_io(png_ptr, fp);

 /* need to delay second part of initialisation until we get the colormap */
 xpix = xpixels;
 ypix = ypixels;
 res_x = (png_uint_32) (xdpi/25.4 + 0.5);
 res_y = (png_uint_32) (ydpi/25.4 + 0.5);
 png_set_pHYs(png_ptr, info_ptr, res_x, res_y, PNG_RESOLUTION_METER);
 return(PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
 int lines, char bandType, unsigned char *cmap)
{
 int x, y;

 if (data==NULL)
  return(PRT_INTERNAL_ERROR);
 if (printModeType<bandType)
  return(PRT_WRONG_IMG);

 /* do not save more data than what was announced in image size */
 if (ypos+lines>ypix)
  lines = min(ypix-ypos, lines);

 if (ypos==0)
  {
   int i;

   switch (printModeType)
    {
     case IMG_BW:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 1, PNG_COLOR_TYPE_GRAY,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      break;
     case IMG_CMAP:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 8, PNG_COLOR_TYPE_PALETTE,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      palette = (png_colorp) png_malloc(png_ptr, 256 * sizeof(png_color));
      for (i = 0;  i < 256;  i++)
       {
	palette[i].red = cmap[i*3];
	palette[i].green = cmap[i*3+1];
	palette[i].blue = cmap[i*3+2];
       }
      png_set_PLTE(png_ptr, info_ptr, palette, 256);
      break;
     case IMG_TRUE_COL:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 8, PNG_COLOR_TYPE_RGB,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      break;
     case IMG_RGBA:
      png_set_IHDR(png_ptr, info_ptr, xpix, ypix, 8, PNG_COLOR_TYPE_RGB_ALPHA,
       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
      break;
    }
   png_write_info(png_ptr, info_ptr);
   png_set_invert_mono(png_ptr);
  }
 
 switch (printModeType)
  {
   case IMG_BW:
    for (y=0; y<lines; y++)
     {
      png_bytep row_pointers = (png_bytep) &data[y*bpl];
      png_write_rows(png_ptr, &row_pointers, 1);
     }
    break;
   case IMG_CMAP:
    if (bandType==IMG_BW)
     {
      unsigned char *dat, cdata[bpl*8];
      png_bytep row_pointers = (png_bytep) &cdata;

      for (y=0; y<lines; y++)
       {
        dat = (unsigned char *) &data[y*bpl];
        for (x=0; x<xpixels; x++)
	 cdata[x] = (unsigned char) ((dat[x>>3]&(128>>(x&7))) ? 255 : 254);
        png_write_rows(png_ptr, &row_pointers, 1);
       }
     }
    else /* (bandType==IMG_CMAP) */
     for (y=0; y<lines; y++)
      {
       png_bytep row_pointers = (png_bytep) &data[y*bpl];
       png_write_rows(png_ptr, &row_pointers, 1);
      }
    break;
   case IMG_TRUE_COL:
    switch (bandType)
     {
      case IMG_BW:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=0; x<xpixels; x++)
	   if (!(dat[x>>3]&(128>>(x&7))))
	    { *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; }
	   else
	    { *cdat++ = 0; *cdat++ = 0; *cdat++ = 0; }
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_CMAP:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=xpixels; x; x--)
	   { *cdat++ = cmap[*dat*3]; *cdat++ = cmap[*dat*3+1]; *cdat++ = cmap[*dat++ *3+2];}
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_TRUE_COL_DITH:
      case IMG_TRUE_COL:
       for (y=0; y<lines; y++)
	{
	 png_bytep row_pointers = (png_bytep) &data[y*bpl];
	 png_write_rows(png_ptr, &row_pointers, 1);
	}
       break;
     }
    break;
   case IMG_RGBA:
    switch (bandType)
     {
      case IMG_BW:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*4];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=0; x<xpixels; x++)
	   if (!(dat[x>>3]&(128>>(x&7))))
	    { *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; }
	   else
	    { *cdat++ = 0; *cdat++ = 0; *cdat++ = 0; *cdat++ = 255; }
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_CMAP:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*4];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=xpixels; x; x--)
	   { *cdat++ = cmap[*dat*3]; *cdat++ = cmap[*dat*3+1]; *cdat++ = cmap[*dat++ *3+2]; *cdat++ = 255; }
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_TRUE_COL_DITH:
      case IMG_TRUE_COL:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*4];
        png_bytep row_pointers = (png_bytep) &cdata;

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=xpixels; x; x--)
	   { *cdat++ = *dat++; *cdat++ = *dat++; *cdat++ = *dat++; *cdat++ = 255; }
          png_write_rows(png_ptr, &row_pointers, 1);
	 }
       }
       break;
      case IMG_RGBA:
       for (y=0; y<lines; y++)
	{
	 png_bytep row_pointers = (png_bytep) &data[y*bpl];
	 png_write_rows(png_ptr, &row_pointers, 1);
	}
       break;
     }
    break;
  } 
 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 if (printStatus==PRT_OK)
  png_write_end(png_ptr, info_ptr);

 if (printModeType==IMG_CMAP)
  {
   png_free(png_ptr, palette);
   palette = NULL;
  }

 png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
 if (fp)
  {
   fclose(fp);
   fp=NULL;
  }
 if ((printStatus==PRT_OK) && fname)
  printStatus = enqueuePrintFile(fname);

 /* just cleanup; called even if printing fails */
 if (fname)
  {
   removeFileFreeFilename(fname, printStatus);
   fname=NULL;
  }
 return(printStatus);
}

/* fill in the preallocated printer modes */
static void getPrtInfo(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_TRUE_COL;
 printerMod[1].printType = IMG_CMAP;
 printerMod[2].printType = IMG_BW;
 printerMod[3].printType = IMG_RGBA;
 *numModes=4;

 *numDensitX = *numDensitY = 7;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = MAX_PRINT;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }
 prtDensityX[6].pixels = prtDensityY[6].pixels = MAX_PRINT;
 prtDensityX[6].dpi = prtDensityY[6].dpi = 720000;

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=TRUE;
 *prtName="PNG";
 *profileName=NULL;
}

/* register PNG "printer" */
void registerPNG()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}
